@@ -1,5 +1,5 @@
 /*
-Copyright (c) 2019 - 2023 Advanced Micro Devices, Inc. All rights reserved.
+Copyright (c) 2019 - 2025 Advanced Micro Devices, Inc. All rights reserved.
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
@@ -24,6 +24,7 @@ THE SOFTWARE.
 // Created by mvx on 3/31/20.
 //
 
+#include <omp.h>
 #include "pipeline/commons.h"
 #include "pipeline/context.h"
 #include "rocal_api.h"
@@ -31,8 +32,7 @@ THE SOFTWARE.
 void
     ROCAL_API_CALL
     rocalRandomBBoxCrop(RocalContext p_context, bool all_boxes_overlap, bool no_crop, RocalFloatParam p_aspect_ratio, bool has_shape, int crop_width, int crop_height, int num_attempts, RocalFloatParam p_scaling, int total_num_attempts, int64_t seed) {
-    if (!p_context)
-        THROW("Invalid rocal context passed to rocalRandomBBoxCrop")
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
     auto context = static_cast<Context*>(p_context);
     FloatParam* aspect_ratio;
     FloatParam* scaling;
@@ -52,8 +52,7 @@ void
 RocalMetaData
     ROCAL_API_CALL
     rocalCreateLabelReader(RocalContext p_context, const char* source_path, const char* file_list_path) {
-    if (!p_context)
-        THROW("Invalid rocal context passed to rocalCreateLabelReader")
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
     auto context = static_cast<Context*>(p_context);
     if (strlen(file_list_path) == 0)
         return context->master_graph->create_label_reader(source_path, MetaDataReaderType::FOLDER_BASED_LABEL_READER);
@@ -64,8 +63,7 @@ RocalMetaData
 RocalMetaData
     ROCAL_API_CALL
     rocalCreateVideoLabelReader(RocalContext p_context, const char* source_path, unsigned sequence_length, unsigned frame_step, unsigned frame_stride, bool file_list_frame_num) {
-    if (!p_context)
-        THROW("Invalid rocal context passed to rocalCreateLabelReader")
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
     auto context = static_cast<Context*>(p_context);
 
     return context->master_graph->create_video_label_reader(source_path, MetaDataReaderType::VIDEO_LABEL_READER, sequence_length, frame_step, frame_stride, file_list_frame_num);
@@ -74,8 +72,7 @@ RocalMetaData
 RocalMetaData
     ROCAL_API_CALL
     rocalCreateCOCOReader(RocalContext p_context, const char* source_path, bool is_output, bool mask, bool ltrb, bool is_box_encoder, bool avoid_class_remapping, bool aspect_ratio_grouping, bool is_box_iou_matcher) {
-    if (!p_context)
-        THROW("Invalid rocal context passed to rocalCreateCOCOReader")
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
     auto context = static_cast<Context*>(p_context);
     if (mask) {
         return context->master_graph->create_coco_meta_data_reader(source_path, is_output, MetaDataReaderType::COCO_META_DATA_READER, MetaDataType::PolygonMask, ltrb, is_box_encoder, avoid_class_remapping, aspect_ratio_grouping, is_box_iou_matcher);
@@ -86,18 +83,16 @@ RocalMetaData
 RocalMetaData
     ROCAL_API_CALL
     rocalCreateCOCOReaderKeyPoints(RocalContext p_context, const char* source_path, bool is_output, float sigma, unsigned pose_output_width, unsigned pose_output_height) {
-    if (!p_context)
-        THROW("Invalid rocal context passed to rocalCreateCOCOReaderKeyPoints")
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
     auto context = static_cast<Context*>(p_context);
 
-    return context->master_graph->create_coco_meta_data_reader(source_path, is_output, MetaDataReaderType::COCO_KEY_POINTS_META_DATA_READER, MetaDataType::KeyPoints, false, false, false, false, sigma, pose_output_width, pose_output_height);
+    return context->master_graph->create_coco_meta_data_reader(source_path, is_output, MetaDataReaderType::COCO_KEY_POINTS_META_DATA_READER, MetaDataType::KeyPoints, false, false, false, false, false, sigma, pose_output_width, pose_output_height);
 }
 
 RocalMetaData
     ROCAL_API_CALL
     rocalCreateTFReader(RocalContext p_context, const char* source_path, bool is_output, const char* user_key_for_label, const char* user_key_for_filename) {
-    if (!p_context)
-        THROW("Invalid rocal context passed to rocalCreateTFReader")
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
     auto context = static_cast<Context*>(p_context);
     std::string user_key_for_label_str(user_key_for_label);
     std::string user_key_for_filename_str(user_key_for_filename);
@@ -114,8 +109,7 @@ RocalMetaData
                                  const char* user_key_for_label, const char* user_key_for_text,
                                  const char* user_key_for_xmin, const char* user_key_for_ymin, const char* user_key_for_xmax, const char* user_key_for_ymax,
                                  const char* user_key_for_filename) {
-    if (!p_context)
-        THROW("Invalid rocal context passed to rocalCreateTFReaderDetection")
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
     auto context = static_cast<Context*>(p_context);
 
     std::string user_key_for_label_str(user_key_for_label);
@@ -141,8 +135,7 @@ RocalMetaData
 RocalMetaData
     ROCAL_API_CALL
     rocalCreateMXNetReader(RocalContext p_context, const char* source_path, bool is_output) {
-    if (!p_context)
-        ERR("Invalid rocal context passed to rocalCreateMXNetReader")
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
     auto context = static_cast<Context*>(p_context);
 
     return context->master_graph->create_mxnet_label_reader(source_path, is_output);
@@ -151,8 +144,7 @@ RocalMetaData
 RocalMetaData
     ROCAL_API_CALL
     rocalCreateTextFileBasedLabelReader(RocalContext p_context, const char* source_path) {
-    if (!p_context)
-        THROW("Invalid rocal context passed to rocalCreateTextFileBasedLabelReader")
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
     auto context = static_cast<Context*>(p_context);
     return context->master_graph->create_label_reader(source_path, MetaDataReaderType::TEXT_FILE_META_DATA_READER);
 }
@@ -160,8 +152,7 @@ RocalMetaData
 void
     ROCAL_API_CALL
     rocalGetImageName(RocalContext p_context, char* buf) {
-    if (!p_context)
-        THROW("Invalid rocal context passed to rocalGetImageName")
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
     auto context = static_cast<Context*>(p_context);
     auto meta_data = context->master_graph->meta_data();
     size_t meta_data_batch_size = meta_data.first.size();
@@ -177,8 +168,7 @@ unsigned
     ROCAL_API_CALL
     rocalGetImageNameLen(RocalContext p_context, int* buf) {
     unsigned size = 0;
-    if (!p_context)
-        THROW("Invalid rocal context passed to rocalGetImageNameLen")
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
     auto context = static_cast<Context*>(p_context);
     auto meta_data = context->master_graph->meta_data();
     size_t meta_data_batch_size = meta_data.first.size();
@@ -194,8 +184,7 @@ unsigned
 void
     ROCAL_API_CALL
     rocalGetImageId(RocalContext p_context, int* buf) {
-    if (!p_context)
-        THROW("Invalid rocal context passed to rocalGetImageId")
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
     auto context = static_cast<Context*>(p_context);
     auto meta_data = context->master_graph->meta_data();
     size_t meta_data_batch_size = meta_data.first.size();
@@ -209,8 +198,7 @@ void
 RocalTensorList
     ROCAL_API_CALL
     rocalGetImageLabels(RocalContext p_context) {
-    if (!p_context)
-        THROW("Invalid rocal context passed to rocalGetImageLabels")
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
     auto context = static_cast<Context*>(p_context);
     return context->master_graph->labels_meta_data();
 }
@@ -218,8 +206,7 @@ RocalTensorList
 unsigned
     ROCAL_API_CALL
     rocalGetBoundingBoxCount(RocalContext p_context) {
-    if (!p_context)
-        THROW("Invalid rocal context passed to rocalGetBoundingBoxCount")
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
     auto context = static_cast<Context*>(p_context);
     auto meta_data = context->master_graph->meta_data();
     if (!meta_data.second)
@@ -233,8 +220,7 @@ unsigned
 RocalTensorList
     ROCAL_API_CALL
     rocalGetBoundingBoxLabel(RocalContext p_context) {
-    if (!p_context)
-        THROW("Invalid rocal context passed to rocalGetBoundingBoxLabel")
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
     auto context = static_cast<Context*>(p_context);
     return context->master_graph->labels_meta_data();
 }
@@ -242,8 +228,7 @@ RocalTensorList
 void
     ROCAL_API_CALL
     rocalGetOneHotImageLabels(RocalContext p_context, void* buf, int num_of_classes, RocalOutputMemType output_mem_type) {
-    if (!p_context)
-        THROW("Invalid rocal context passed to rocalGetOneHotImageLabels")
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
     auto context = static_cast<Context*>(p_context);
     auto meta_data = context->master_graph->meta_data();
     if (!meta_data.second) {
@@ -282,12 +267,20 @@ void
 RocalTensorList
     ROCAL_API_CALL
     rocalGetBoundingBoxCords(RocalContext p_context) {
-    if (!p_context)
-        THROW("Invalid rocal context passed to rocalGetBoundingBoxCords")
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
     auto context = static_cast<Context*>(p_context);
     return context->master_graph->bbox_meta_data();
 }
 
+RocalMetaData
+    ROCAL_API_CALL
+    rocalGetAsciiDatas(RocalContext p_context) {
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
+    auto context = static_cast<Context*>(p_context);
+    return context->master_graph->ascii_values_meta_data();
+}
+
+
 unsigned
     ROCAL_API_CALL
     rocalGetMaskCount(RocalContext p_context, int* buf) {
@@ -341,8 +334,7 @@ RocalTensorList
 void
     ROCAL_API_CALL
     rocalGetImageSizes(RocalContext p_context, int* buf) {
-    if (!p_context)
-        THROW("Invalid rocal context passed to rocalGetImageSizes")
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
     auto context = static_cast<Context*>(p_context);
     auto meta_data = context->master_graph->meta_data();
     size_t meta_data_batch_size = meta_data.second->get_img_sizes_batch().size();
@@ -360,10 +352,7 @@ void
 void
     ROCAL_API_CALL
     rocalGetROIImageSizes(RocalContext p_context, int* buf) {
-    if (!p_context) {
-        THROW("Invalid rocal context passed to rocalGetROIImageSizes")
-        return;
-    }
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
     auto context = static_cast<Context*>(p_context);
     try {
         auto meta_data = context->master_graph->meta_data();
@@ -386,8 +375,7 @@ void
 RocalMetaData
     ROCAL_API_CALL
     rocalCreateTextCifar10LabelReader(RocalContext p_context, const char* source_path, const char* file_prefix) {
-    if (!p_context)
-        THROW("Invalid rocal context passed to rocalCreateTextCifar10LabelReader")
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
     auto context = static_cast<Context*>(p_context);
 
     return context->master_graph->create_cifar10_label_reader(source_path, file_prefix);
@@ -396,8 +384,7 @@ RocalMetaData
 void
     ROCAL_API_CALL
     rocalGetSequenceStartFrameNumber(RocalContext p_context, unsigned int* buf) {
-    if (!p_context)
-        THROW("Invalid rocal context passed to rocalGetSequenceStartFrameNumber")
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
     auto context = static_cast<Context*>(p_context);
     std::vector<size_t> sequence_start_frame;
     context->master_graph->sequence_start_frame_number(sequence_start_frame);
@@ -407,8 +394,7 @@ void
 void
     ROCAL_API_CALL
     rocalGetSequenceFrameTimestamps(RocalContext p_context, float* buf) {
-    if (!p_context)
-        THROW("Invalid rocal context passed to rocalGetSequenceFrameTimestamps")
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
     auto context = static_cast<Context*>(p_context);
     std::vector<std::vector<float>> sequence_frame_timestamps;
     context->master_graph->sequence_frame_timestamps(sequence_frame_timestamps);
@@ -423,8 +409,7 @@ void
     ROCAL_API_CALL
     rocalBoxEncoder(RocalContext p_context, std::vector<float>& anchors, float criteria,
                     std::vector<float>& means, std::vector<float>& stds, bool offset, float scale) {
-    if (!p_context)
-        THROW("Invalid rocal context passed to rocalBoxEncoder")
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
     auto context = static_cast<Context*>(p_context);
     context->master_graph->box_encoder(anchors, criteria, means, stds, offset, scale);
 }
@@ -432,8 +417,7 @@ void
 void
     ROCAL_API_CALL
     rocalCopyEncodedBoxesAndLables(RocalContext p_context, float* boxes_buf, int* labels_buf) {
-    if (!p_context)
-        THROW("Invalid rocal context passed to rocalCopyEncodedBoxesAndLables")
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
     auto context = static_cast<Context*>(p_context);
     auto meta_data = context->master_graph->meta_data();
     size_t meta_data_batch_size = meta_data.second->get_labels_batch().size();
@@ -463,9 +447,7 @@ void
 RocalMetaData
     ROCAL_API_CALL
     rocalGetEncodedBoxesAndLables(RocalContext p_context, int num_encoded_boxes) {
-    if (!p_context) {
-        WRN("rocalGetEncodedBoxesAndLables::Invalid context")
-    }
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
     auto context = static_cast<Context*>(p_context);
     return context->master_graph->get_bbox_encoded_buffers(num_encoded_boxes);
 }
@@ -473,8 +455,7 @@ RocalMetaData
 void
     ROCAL_API_CALL
     rocalGetJointsDataPtr(RocalContext p_context, RocalJointsData** joints_data) {
-    if (!p_context)
-        THROW("Invalid rocal context passed to rocalGetBoundingBoxCords")
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
     auto context = static_cast<Context*>(p_context);
     auto meta_data = context->master_graph->meta_data();
     size_t meta_data_batch_size = meta_data.second->get_joints_data_batch().center_batch.size();
@@ -495,19 +476,29 @@ void
                        std::vector<float>& anchors,
                        float high_threshold, float low_threshold,
                        bool allow_low_quality_matches) {
-    if (!p_context)
-        THROW("Invalid rocal context passed to rocalBoxIouMatcher")
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
     auto context = static_cast<Context*>(p_context);
     context->master_graph->box_iou_matcher(anchors, high_threshold,
                                            low_threshold,
                                            allow_low_quality_matches);
 }
 
+RocalMetaData
+    ROCAL_API_CALL
+    rocalCreateWebDatasetReader(RocalContext p_context, const char* source_path, const char* index_path, std::vector<std::set<std::string>> extensions, RocalMissingComponentsBehaviour missing_components_behavior, bool is_output) {
+#ifdef ENABLE_WDS
+        ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
+        auto context = static_cast<Context*>(p_context);
+        return context->master_graph->create_webdataset_reader(source_path, index_path, extensions , MetaDataReaderType::WEBDATASET_META_DATA_READER, static_cast<MissingComponentsBehaviour>(missing_components_behavior));
+#else
+        THROW("Webdataset reader is not enabled since libtar is not present")
+#endif
+}
+
 RocalTensorList
     ROCAL_API_CALL
     rocalGetMatchedIndices(RocalContext p_context) {
-    if (!p_context)
-        THROW("Invalid rocal context passed to rocalGetMatchedIndices")
+    ROCAL_INVALID_CONTEXT_EXCEPTION(p_context);
     auto context = static_cast<Context*>(p_context);
     return context->master_graph->matched_index_meta_data();
 }
