@@ -1,5 +1,5 @@
 /*
-Copyright (c) 2019 - 2023 Advanced Micro Devices, Inc. All rights reserved.
+Copyright (c) 2019 - 2025 Advanced Micro Devices, Inc. All rights reserved.
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
@@ -22,6 +22,7 @@ THE SOFTWARE.
 #if ENABLE_OPENCL
 #include <CL/cl.h>
 #endif
+#include <omp.h>
 #include <vx_ext_amd.h>
 #include <VX/vx_types.h>
 #include <cstring>
@@ -259,6 +260,26 @@ void MasterGraph::create_single_graph() {
     _graph->verify();
 }
 
+void MasterGraph::create_multiple_graphs() {
+    // Actual graph creating and calls into adding nodes to graph is deferred and is happening here to enable potential future optimizations
+    // Creating a Graph instance for every loader module in the pipeline
+    for (unsigned n = 0; n < _loaders_count; n++) {
+        _graphs.emplace_back(std::make_shared<Graph>(_context, _affinity, 0, _cpu_num_threads, _gpu_id));
+    }
+    for (auto &node : _nodes) {
+        // Any tensor not yet created can be created as virtual tensor
+        for (auto &tensor : node->output())
+            if (tensor->info().type() == TensorInfo::Type::UNKNOWN) {
+                tensor->create_virtual(_context, _graphs[node->get_graph_id()]->get());
+                _internal_tensors.push_back(tensor);
+            }
+        node->create(_graphs[node->get_graph_id()]);
+    }
+
+    for (auto &graph : _graphs)
+        graph->verify();
+}
+
 MasterGraph::Status
 MasterGraph::build() {
     if (_internal_tensor_list.empty())
@@ -270,15 +291,26 @@ MasterGraph::build() {
     _ring_buffer.init(_mem_type, nullptr, _internal_tensor_list.data_size(), _internal_tensor_list.roi_size());
 #endif
     if (_is_box_encoder) _ring_buffer.initBoxEncoderMetaData(_mem_type, _user_batch_size * _num_anchors * 4 * sizeof(float), _user_batch_size * _num_anchors * sizeof(int));
-    create_single_graph();
+    
+    // Check if at least one loader module is created
+    if (_loader_modules.size() < 1)
+        THROW("At least one loader needs to be created in the pipeline")
+
+    if (_loaders_count > 1) {
+        _meta_data_reader = nullptr; // Disable metadata reader for multiple loaders pipeline, support not enabled
+        create_multiple_graphs();
+    } else {
+        _loader_module = _loader_modules[0];
+        create_single_graph();
+    }
     start_processing();
     return Status::OK;
 }
 
 Tensor *
-MasterGraph::create_loader_output_tensor(const TensorInfo &info) {
+MasterGraph::create_internal_tensor(const TensorInfo &info) {
     /*
-     *   NOTE: Output tensor for a source node needs to be created as a regular (non-virtual) tensor
+     *   NOTE: This function creates a regular (non-virtual) tensor
      */
     auto output = new Tensor(info);
     if (output->create_from_handle(_context) != 0)
@@ -320,13 +352,18 @@ void MasterGraph::set_output(Tensor *output_tensor) {
 void MasterGraph::release() {
     LOG("MasterGraph release ...")
     stop_processing();
+    for (auto &node : _nodes)
+        node->release();
     _nodes.clear();
+    for (auto &node : _root_nodes)
+        node->release();
     _root_nodes.clear();
     _meta_data_nodes.clear();
     _tensor_map.clear();
     _ring_buffer.release_gpu_res();
     // shut_down loader:: required for releasing any allocated resourses
-    _loader_module->shut_down();
+    for (auto &loader_module : _loader_modules)
+        loader_module->shut_down();
     // release output buffer if allocated
     if (_output_tensor_buffer != nullptr) {
 #if ENABLE_OPENCL
@@ -347,16 +384,21 @@ void MasterGraph::release() {
     _internal_tensor_list.release();  // It will call the vxReleaseTensor internally in the destructor for each tensor in the list
     _output_tensor_list.release();    // It will call the vxReleaseTensor internally in the destructor for each tensor in the list
     _metadata_output_tensor_list.release(); // It will call the vxReleaseTensor internally in the destructor for each tensor in the list of TensorList
-    _bbox_encoded_output.release(); // It will call the vxReleaseTensor internally in the destructor for each tensor in the list of TensorList
 
     if (_graph != nullptr)
         _graph->release();
+
+    // Release the graph created for each loader module, in case of multiple loaders
+    for (auto& graph : _graphs) {
+        graph->release();
+    }
     if (_meta_data_reader != nullptr)
         _meta_data_reader->release();
 
     _augmented_meta_data = nullptr;
     _meta_data_graph = nullptr;
     _meta_data_reader = nullptr;
+    delete _box_encoder_gpu;
     if (_context && (status = vxReleaseContext(&_context)) != VX_SUCCESS)
         LOG("Failed to call vxReleaseContext " + TOSTR(status))
 }
@@ -418,7 +460,8 @@ MasterGraph::reset() {
     if (_randombboxcrop_meta_data_reader != nullptr)
         _randombboxcrop_meta_data_reader->release();
     // resetting loader module to start from the beginning of the media and clear it's internal state/buffers
-    _loader_module->reset();
+    for (auto &loader_module : _loader_modules)
+        loader_module->reset();
     // restart processing of the images
     _first_run = true;
     _output_routine_finished_processing = false;
@@ -440,12 +483,22 @@ MasterGraph::mem_type() {
 
 size_t
 MasterGraph::last_batch_padded_size() {
-    return _loader_module->last_batch_padded_size();
+    size_t max_last_batch_padded_size = 0;
+    for (auto loader_module : _loader_modules)
+        max_last_batch_padded_size = std::max(loader_module->last_batch_padded_size(), max_last_batch_padded_size);
+    return max_last_batch_padded_size;
 }
 
 Timing
 MasterGraph::timing() {
-    Timing t = _loader_module->timing();
+    Timing t;
+    // Accumulate the timings from each loader
+    for (auto loader_module : _loader_modules) {
+        Timing loader_time = loader_module->timing();
+        t.decode_time += loader_time.decode_time;
+        t.read_time += loader_time.read_time;
+        t.process_time += loader_time.process_time;
+    }
     t.process_time += _process_time.get_timing();
     t.copy_to_output += _convert_time.get_timing();
     t.bb_process_time += _bencode_time.get_timing();
@@ -885,6 +938,16 @@ MasterGraph::get_output_tensors() {
     return &_output_tensor_list;
 }
 
+bool MasterGraph::is_out_of_data() {
+    // If any of the loader module's remaining count is less than the batch size, return loader out of data
+    for (auto& loader_module : _loader_modules) {
+        if (loader_module->remaining_count() < (_is_sequence_reader_output ? _sequence_batch_size : _user_batch_size)) {
+            return true;
+        }
+    }
+    return false;
+}
+
 void MasterGraph::output_routine() {
     INFO("Output routine started with " + TOSTR(_remaining_count) + " to load");
     try {
@@ -987,10 +1050,74 @@ void MasterGraph::output_routine() {
     }
 }
 
+void MasterGraph::output_routine_multiple_loaders() {
+    INFO("Output routine for multiple loaders started with " + TOSTR(_remaining_count) + " to load");
+    try {
+        while (_processing) {
+            if (is_out_of_data()) {
+                // If the internal process routine ,output_routine(), has finished processing all the images, and last
+                // processed images stored in the _ring_buffer will be consumed by the user when it calls the run() func
+                notify_user_thread();
+                // the following call is required in case the ring buffer is waiting for more data to be loaded and there is no more data to process.
+                _ring_buffer.release_if_empty();
+                std::this_thread::sleep_for(std::chrono::milliseconds(100));
+                continue;
+            }
+            _rb_block_if_full_time.start();
+            // _ring_buffer.get_write_buffers() is blocking and blocks here until user uses processed image by calling run() and frees space in the ring_buffer
+            auto write_buffers = _ring_buffer.get_write_buffers();
+            auto write_output_buffers = write_buffers.first;
+            _rb_block_if_full_time.end();
+
+            // Swap handles on the input tensor, so that new tensor is loaded to be processed
+            for (auto loader_module : _loader_modules) {
+                auto load_ret = loader_module->load_next();
+                if (load_ret != LoaderModuleStatus::OK)
+                    THROW("Loader module failed to load next batch of images, status " + TOSTR(load_ret))
+            }
+
+            if (!_processing)
+                break;
+
+            // Swap handles on the output tensor, so that new processed tensor will be written to the a new buffer
+            for (size_t idx = 0; idx < _internal_tensor_list.size(); idx++)
+                _internal_tensor_list[idx]->swap_handle(write_output_buffers[idx]);
+
+            if (!_processing)
+                break;
+
+            update_node_parameters();
+            _process_time.start();
+            for (auto& graph : _graphs) {
+                graph->process();
+            }
+            _process_time.end();
+
+            auto write_roi_buffers = write_buffers.second;   // Obtain ROI buffers from ring buffer
+            for (size_t idx = 0; idx < _internal_tensor_list.size(); idx++)
+                _internal_tensor_list[idx]->copy_roi(write_roi_buffers[idx]);   // Copy ROI from internal tensor's buffer to ring buffer
+
+            _ring_buffer.push();  // Image data and metadata is now stored in output the ring_buffer, increases it's level by 1
+        }
+    } catch (const std::exception &e) {
+        ERR("Exception thrown in the process routine: " + STR(e.what()) + STR("\n"));
+        _processing = false;
+        _ring_buffer.release_all_blocked_calls();
+    }
+}
+
 void MasterGraph::start_processing() {
     _processing = true;
-    _remaining_count = _loader_module->remaining_count();
-    _output_thread = std::thread(&MasterGraph::output_routine, this);
+    _remaining_count = _loader_modules[0]->remaining_count();
+    for (int i = 1; i < _loaders_count; i++) {
+        // Stores the least remaining count value of all loaders
+        _remaining_count = std::min(_remaining_count, static_cast<int>(_loader_modules[i]->remaining_count()));
+    }
+    if (_loaders_count == 1) {
+        _output_thread = std::thread(&MasterGraph::output_routine, this);
+    } else {
+        _output_thread = std::thread(&MasterGraph::output_routine_multiple_loaders, this);
+    }
 #if defined(WIN32) || defined(_WIN32) || defined(__WIN32) && !defined(__CYGWIN__)
 #else
 //  Changing thread scheduling policy and it's priority does not help on latest Ubuntu builds
@@ -1137,8 +1264,7 @@ TensorListVector* MasterGraph::create_label_reader(const char *source_path, Meta
         THROW("A metadata reader has already been created")
     if (_augmented_meta_data)
         THROW("Metadata can only have a single output")
-    if (strlen(source_path) == 0)
-        THROW("Source path needs to be provided")
+
     MetaDataConfig config(MetaDataType::Label, reader_type, source_path);
     _meta_data_reader = create_meta_data_reader(config, _augmented_meta_data);
     _meta_data_reader->read_all(source_path);
@@ -1158,6 +1284,46 @@ TensorListVector* MasterGraph::create_label_reader(const char *source_path, Meta
     return &_metadata_output_tensor_list;
 }
 
+TensorListVector* MasterGraph::create_webdataset_reader(
+    const char *source_path, const char *index_path,
+    std::vector<std::set<std::string>> extensions,
+    MetaDataReaderType reader_type, MissingComponentsBehaviour missing_component_behaviour) {
+    if (_meta_data_reader)
+        THROW("A metadata reader has already been created")
+    if (_augmented_meta_data)
+        THROW("Metadata can only have a single output")
+
+    bool generate_index = (index_path[0] == '\0') ? true : false;
+    if (generate_index)
+        std::cerr << "Index file is not provided, it may take some time to infer it from the tar file";
+
+    _ascii_tensor_list.resize(extensions[0].size() - 1);
+    MetaDataConfig config(MetaDataType::AsciiValue, reader_type, source_path,
+                          std::map<std::string, std::string>(), std::string(),
+                          0, 0, 0, index_path, missing_component_behaviour,
+                          extensions);
+    _meta_data_reader = create_meta_data_reader(config, _augmented_meta_data);
+    _meta_data_reader->read_all(source_path);
+    std::vector<size_t> dims;
+    dims = {MAX_ASCII_BUFFER};
+    auto default_ascii_values_info = TensorInfo(std::move(dims), _mem_type,RocalTensorDataType::UINT8); // Create default ascii values Info
+    default_ascii_values_info.set_metadata();
+    for (uint ext_count = 0; ext_count < _ascii_tensor_list.size(); ext_count++) {
+        for (unsigned i = 0; i < _user_batch_size; i++) {
+            _meta_data_buffer_size.emplace_back(_user_batch_size * default_ascii_values_info.data_size());
+            auto info = default_ascii_values_info;
+            auto tensor = new Tensor(info);
+            _ascii_tensor_list[ext_count].push_back(tensor);
+        }
+        _metadata_output_tensor_list.emplace_back(&_ascii_tensor_list[ext_count]);
+        _webdataset_output_tensor_list.emplace_back(&_ascii_tensor_list[ext_count]);
+    }
+
+    _ring_buffer.init_metadata(RocalMemType::HOST, _meta_data_buffer_size);
+
+    return &_metadata_output_tensor_list;
+}
+
 TensorListVector* MasterGraph::create_video_label_reader(const char *source_path, MetaDataReaderType reader_type, unsigned sequence_length, unsigned frame_step, unsigned frame_stride, bool file_list_frame_num) {
     if (_meta_data_reader)
         THROW("A metadata reader has already been created")
@@ -1366,8 +1532,10 @@ TensorListVector* MasterGraph::create_cifar10_label_reader(const char *source_pa
 }
 
 const std::pair<ImageNameBatch, pMetaDataBatch> &MasterGraph::meta_data() {
-    if (_ring_buffer.level() == 0)
+    if (_ring_buffer.level() == 0) {
+        release();
         THROW("No meta data has been loaded")
+    }
     return _ring_buffer.get_meta_data();
 }
 
@@ -1394,8 +1562,14 @@ size_t MasterGraph::bounding_box_batch_count(pMetaDataBatch meta_data_batch) {
 TensorList *MasterGraph::labels_meta_data() {
     if (_external_source_reader)
         return &_labels_tensor_list;
-    if (_ring_buffer.level() == 0)
+
+    if (!_meta_data_reader && _loaders_count > 1)
+        THROW("Metadata reader is not compatible with multiple loaders")
+
+    if (_ring_buffer.level() == 0) {
+        release();
         THROW("No meta data has been loaded")
+    }
     auto meta_data_buffers = (unsigned char *)_ring_buffer.get_meta_read_buffers()[0];  // Get labels buffer from ring buffer
     auto labels = _ring_buffer.get_meta_data().second->get_labels_batch();
     for (unsigned i = 0; i < _labels_tensor_list.size(); i++) {
@@ -1406,9 +1580,43 @@ TensorList *MasterGraph::labels_meta_data() {
     return &_labels_tensor_list;
 }
 
-TensorList *MasterGraph::bbox_meta_data() {
-    if (_ring_buffer.level() == 0)
+TensorListVector *MasterGraph::ascii_values_meta_data() {
+    if (!_meta_data_reader && _loaders_count > 1)
+        THROW("Metadata reader is not compatible with multiple loaders")
+
+    if (_external_source_reader) {
+        return &_webdataset_output_tensor_list;
+    }
+    if (_ring_buffer.level() == 0) { 
+        release();
         THROW("No meta data has been loaded")
+    }
+
+    for (uint ext = 0; ext < _ascii_tensor_list.size(); ext++) {
+        auto meta_data_buffers = (uint8_t *)_ring_buffer.get_meta_read_buffers()[ext]; // Get ASCII buffer from ring buffer
+        auto ascii_values = _ring_buffer.get_meta_data().second->get_ascii_values_batch();
+        for (unsigned i = 0; i < _ascii_tensor_list[ext].size(); i++) {
+            if (ascii_values[i][ext]) {
+                _ascii_tensor_list[ext][i]->set_dims({ascii_values[i][ext]->size()});
+                _ascii_tensor_list[ext][i]->set_mem_handle((void *)meta_data_buffers);
+                meta_data_buffers += _ascii_tensor_list[ext][i]->info().data_size();
+            } else { // To Handle Empty Case of Missing Behaviour Component
+                _ascii_tensor_list[ext][i]->set_dims({0});
+                _ascii_tensor_list[ext][i]->reset_mem_handle();
+                meta_data_buffers += _ascii_tensor_list[ext][i]->info().data_size();
+            }
+        }
+    }
+    return &_webdataset_output_tensor_list;
+}
+
+TensorList *MasterGraph::bbox_meta_data() {
+    if (!_meta_data_reader && _loaders_count > 1)
+        THROW("Metadata reader is not compatible with multiple loaders")
+    if (_ring_buffer.level() == 0) {
+        release();
+        THROW("No meta data has been loaded")
+    }
     auto meta_data_buffers = (unsigned char *)_ring_buffer.get_meta_read_buffers()[1];  // Get bbox buffer from ring buffer
     auto bbox_cords = _ring_buffer.get_meta_data().second->get_bb_cords_batch();
     for (unsigned i = 0; i < _bbox_tensor_list.size(); i++) {
@@ -1421,8 +1629,12 @@ TensorList *MasterGraph::bbox_meta_data() {
 }
 
 TensorList *MasterGraph::mask_meta_data() {
-    if (_ring_buffer.level() == 0)
+    if (!_meta_data_reader && _loaders_count > 1)
+        THROW("Metadata reader is not compatible with multiple loaders")
+    if (_ring_buffer.level() == 0) {
+        release();
         THROW("No meta data has been loaded")
+    }
     auto meta_data_buffers = (unsigned char *)_ring_buffer.get_meta_read_buffers()[2];  // Get mask buffer from ring buffer
     auto mask_cords = _ring_buffer.get_meta_data().second->get_mask_cords_batch();
     for (unsigned i = 0; i < _mask_tensor_list.size(); i++) {
@@ -1435,8 +1647,13 @@ TensorList *MasterGraph::mask_meta_data() {
 }
 
 TensorList *MasterGraph::matched_index_meta_data() {
-    if (_ring_buffer.level() == 0)
+    if (!_meta_data_reader && _loaders_count > 1)
+        THROW("Metadata reader is not compatible with multiple loaders")
+
+    if (_ring_buffer.level() == 0) {
+        release();
         THROW("No meta data has been loaded")
+    }
     auto meta_data_buffers = reinterpret_cast<unsigned char *>(_ring_buffer.get_meta_read_buffers()[2]);  // Get matches buffer from ring buffer
     for (unsigned i = 0; i < _matches_tensor_list.size(); i++) {
         _matches_tensor_list[i]->set_mem_handle(reinterpret_cast<void *>(meta_data_buffers));
@@ -1465,17 +1682,17 @@ MasterGraph::copy_out_tensor_planar(void *out_ptr, RocalTensorlayout format, flo
     _convert_time.start();
     // Copies to the output context given by the user, each image is copied separate for planar
     auto output_tensor_info = _output_tensor_list[0]->info();
-    auto dims = output_tensor_info.dims();
-    const size_t w = dims[2];
-    const size_t h = dims[1];
-    const size_t c = dims[3];
-    const size_t n = dims[0];
 
     const size_t single_output_tensor_size = output_tensor_info.data_size();
 
     if (output_tensor_info.mem_type() == RocalMemType::OCL || output_tensor_info.mem_type() == RocalMemType::HIP) {
         THROW("copy_out_tensor_planar for GPU affinity is not implemented")
     } else if (output_tensor_info.mem_type() == RocalMemType::HOST) {
+        auto dims = output_tensor_info.dims();
+        const size_t n = dims[0];
+        const size_t c = dims[1];
+        const size_t h = dims[2];
+        const size_t w = dims[3];
         float multiplier[3] = {multiplier0, multiplier1, multiplier2};
         float offset[3] = {offset0, offset1, offset2};
         size_t dest_buf_offset = 0;
@@ -1606,6 +1823,9 @@ MasterGraph::copy_out_tensor_planar(void *out_ptr, RocalTensorlayout format, flo
 
 TensorListVector*
 MasterGraph::get_bbox_encoded_buffers(size_t num_encoded_boxes) {
+    if (!_meta_data_reader && _loaders_count > 1)
+        THROW("Metadata reader is not compatible with multiple loaders")
+
     if (_is_box_encoder) {
         if (num_encoded_boxes != _user_batch_size * _num_anchors) {
             THROW("num_encoded_boxes is not correct");
@@ -1613,20 +1833,21 @@ MasterGraph::get_bbox_encoded_buffers(size_t num_encoded_boxes) {
         auto encoded_boxes_and_lables = _ring_buffer.get_box_encode_read_buffers();
         unsigned char *boxes_buf_ptr = (unsigned char *)encoded_boxes_and_lables.first;
         unsigned char *labels_buf_ptr = (unsigned char *)encoded_boxes_and_lables.second;
-        auto labels = _ring_buffer.get_meta_data().second->get_labels_batch();
 
         if (_bbox_tensor_list.size() != _labels_tensor_list.size())
             THROW("The number of tensors between bbox and bbox_labels do not match")
         for (unsigned i = 0; i < _bbox_tensor_list.size(); i++) {
-            _labels_tensor_list[i]->set_dims({labels[i].size()});
-            _bbox_tensor_list[i]->set_dims({labels[i].size(), 4});
             _labels_tensor_list[i]->set_mem_handle((void *)labels_buf_ptr);
             _bbox_tensor_list[i]->set_mem_handle((void *)boxes_buf_ptr);
             labels_buf_ptr += _labels_tensor_list[i]->info().data_size();
             boxes_buf_ptr += _bbox_tensor_list[i]->info().data_size();
         }
-        _bbox_encoded_output.emplace_back(&_labels_tensor_list);
-        _bbox_encoded_output.emplace_back(&_bbox_tensor_list);
+
+        // Set the labels and bbox tensorList to the box encoded output only for the first run
+        if (_bbox_encoded_output.size() == 0) {
+            _bbox_encoded_output.emplace_back(&_labels_tensor_list);
+            _bbox_encoded_output.emplace_back(&_bbox_tensor_list);
+        }
     }
     return &_bbox_encoded_output;
 }
@@ -1635,6 +1856,8 @@ void MasterGraph::feed_external_input(const std::vector<std::string>& input_imag
                                       const std::vector<ROIxywh>& roi_xywh, unsigned int max_width, unsigned int max_height, unsigned int channels,
                                       ExternalSourceFileMode mode, RocalTensorlayout layout, bool eos) {
     _external_source_eos = eos;
+    if (!_loader_module)
+        THROW("Loader module does not exist")
     _loader_module->feed_external_input(input_images_names, input_buffer, roi_xywh, max_width, max_height, channels, mode, eos);
 
     if (is_labels) {
